import warnings
warnings.simplefilter(action="ignore", category=UserWarning)
import h2o
import lightgbm
import catboost
import xgboost
from tsforest.config import cat_parameters, lgbm_parameters, gbm_parameters, xgb_parameters
from tsforest.forest_base import BaseRegressor

NTREES_LGB_ALIASES = {"num_iteration", "num_iterations", 
                      "n_iter", "num_tree", "num_trees", 
                      "num_round", "num_rounds"}
EARLY_STOP_LGB_ALIASES = {"early_stopping_round",
                          "early_stopping_rounds",
                          "early_stopping",
                          "n_iter_no_change"}

def fetch_param_key(model_params, aliases):
    keys = model_params.keys() & aliases
    assert len(keys) <= 1, \
        f"Same parameter with different aliases in 'model_params': {keys}."
    if len(keys)==0: return None
    else: return list(keys)[0]

class H2OGBMRegressor(BaseRegressor):
    def __init__(self, model_params):
        self.model_params = {**gbm_parameters, **model_params}
        self.model = None
    
    def set_params(self, model_params):
        self.model_params = {**gbm_parameters, **model_params}
    
    def get_params(self):
        return self.model_params

    def cast_dataframe(self, features_dataframe, input_features, target,  categorical_features):
        features_types = {feature:"categorical" for feature in categorical_features
                          if feature in input_features}
        features_dataframe_casted = h2o.H2OFrame(features_dataframe, 
                                                 column_types=features_types)
        return features_dataframe_casted
    
    def fit(self, train_features, valid_features, input_features, target, categorical_features, fit_kwargs=dict()):
        train_features_casted = self.cast_dataframe(train_features, input_features, target, categorical_features)
        valid_features_casted = self.cast_dataframe(valid_features, input_features, target, categorical_features) \
                                if valid_features is not None else None
        model_params = dict(self.model_params)
        training_params = {"training_frame":train_features_casted, 
                           "x":input_features, 
                           "y":target}
        if valid_features is not None:
            training_params["validation_frame"] = valid_features_casted
        elif "stopping_rounds" in model_params:
            del model_params["stopping_rounds"]
        if "weight" in train_features.columns:
            training_params["weights_column"] = "weight"
        # model training
        self.model = h2o.estimators.H2OGradientBoostingEstimator(**model_params)
        self.model.train(**training_params, **fit_kwargs)
        self.best_iteration = int(self.model.summary()["number_of_trees"][0])
        self.input_features = input_features
        self.target = target
        self.categorical_features = categorical_features

    def predict(self, predict_features):
        predict_features_casted = self.cast_dataframe(predict_features, self.input_features, self.target, self.categorical_features)
        _prediction = self.model.predict(predict_features_casted)
        prediction = _prediction.as_data_frame().values[:,0]
        return prediction

    def save_model(self, fname, **kwargs):
        if self.model is not None:
            h2o.save_model(model=self.model, path=fname, **kwargs)
    
    def load_model(self, fname, **kwargs):
        self.model = h2o.load_model(fname, **kwargs)


class LightGBMRegressor(BaseRegressor):
    def __init__(self, model_params):
        self.model_params = {**lgbm_parameters, **model_params}
        self.model = None

    def set_params(self, model_params):
        self.model_params = {**lgbm_parameters, **model_params}
    
    def get_params(self):
        return self.model_params

    def cast_dataframe(self, features_dataframe, input_features, target, categorical_features):
        dataset_params = {"data":features_dataframe.loc[:, input_features],
                          "categorical_feature":categorical_features,
                          "free_raw_data":False}
        if "weight" in features_dataframe.columns:
            dataset_params["weight"] = features_dataframe.loc[:, "weight"].values
        if target in features_dataframe.columns:
            dataset_params["label"] = features_dataframe.loc[:, target].values
        features_dataframe_casted = lightgbm.Dataset(**dataset_params)
        return features_dataframe_casted 
    
    def fit(self, train_features, valid_features, input_features, target, categorical_features, fit_kwargs=dict()):
        train_features_casted = self.cast_dataframe(train_features, input_features, target, categorical_features)
        valid_features_casted = self.cast_dataframe(valid_features, input_features, target, categorical_features) \
                                if valid_features is not None else None
        model_params = dict(self.model_params)
        ntrees_param_key = fetch_param_key(model_params, NTREES_LGB_ALIASES)
        early_stop_param_key = fetch_param_key(model_params, EARLY_STOP_LGB_ALIASES)
        training_params = {"train_set":train_features_casted}
        if valid_features is not None:
            training_params["valid_sets"] = valid_features_casted
            if early_stop_param_key is not None:
                training_params["early_stopping_rounds"] = model_params.pop(early_stop_param_key)
        elif early_stop_param_key is not None:
            del model_params[early_stop_param_key]
        if ntrees_param_key is not None:
            training_params["num_boost_round"] = model_params.pop(ntrees_param_key) 
        training_params["params"] = model_params
        # model training
        self.model = lightgbm.train(**training_params, **fit_kwargs)
        self.best_iteration = self.model.best_iteration if self.model.best_iteration>0 else self.model.num_trees()
        self.input_features = input_features
        self.target = target
        self.categorical_features = categorical_features

    def predict(self, predict_features):
        prediction = self.model.predict(predict_features.loc[:, self.input_features])
        return prediction
    
    def save_model(self, fname, **kwargs):
        if self.model is not None:
            self.model.save_model(fname, **kwargs)
    
    def load_model(self, fname, **kwargs):
        self.model = lightgbm.Booster(model_file=fname, **kwargs)


class CatBoostRegressor(BaseRegressor):
    def __init__(self, model_params):
        self.model_params = {**cat_parameters, **model_params}
        self.model = None

    def set_params(self, model_params):
        self.model_params = {**cat_parameters, **model_params}
    
    def get_params(self):
        return self.model_params

    def cast_dataframe(self, features_dataframe, input_features, target,  categorical_features):
        dataset_params = {"data":features_dataframe.loc[:, input_features],
                          "cat_features":categorical_features}
        if "weight" in features_dataframe.columns:
            dataset_params["weight"] = features_dataframe.loc[:, "weight"].values
        if target in features_dataframe.columns:
            dataset_params["label"] = features_dataframe.loc[:, target].values
        features_dataframe_casted = catboost.Pool(**dataset_params)
        return features_dataframe_casted
    
    def fit(self, train_features, valid_features, input_features, target, categorical_features, fit_kwargs=dict()):
        train_features_casted = self.cast_dataframe(train_features, input_features, target, categorical_features)
        valid_features_casted = self.cast_dataframe(valid_features, input_features, target, categorical_features) \
                                if valid_features is not None else None
        model_params = dict(self.model_params)
        training_params = {"X":train_features_casted}
        if valid_features is not None:
            training_params["eval_set"] = valid_features_casted
        elif "early_stopping_rounds" in model_params:
            del model_params["early_stopping_rounds"]
        # model training
        self.model = catboost.CatBoostRegressor(**model_params)
        self.model.fit(**training_params, **fit_kwargs)
        self.best_iteration = self.model.best_iteration_ if self.model.best_iteration_ is not None else self.model.tree_count_
        self.input_features = input_features
        self.target = target
        self.categorical_features = categorical_features
    
    def predict(self, predict_features):
        predict_features_casted = self.cast_dataframe(predict_features, self.input_features, self.target, self.categorical_features)
        prediction = self.model.predict(predict_features_casted)
        return prediction

    def save_model(self, fname, **kwargs):
        if self.model is not None:
            self.model.save_model(fname, **kwargs)
    
    def load_model(self, fname, **kwargs):
        self.model = catboost.CatBoostRegressor()
        self.model.load_model(fname, **kwargs)


class XGBoostRegressor(BaseRegressor):
    def __init__(self, model_params):
        self.model_params = {**xgb_parameters, **model_params}
        self.model = None

    def set_params(self, model_params):
        self.model_params = {**xgb_parameters, **model_params}
    
    def get_params(self):
        return self.model_params

    def cast_dataframe(self, features_dataframe, input_features, target,  categorical_features):
        dataset_params = {"data":features_dataframe.loc[:, input_features]}
        if "weight" in features_dataframe.columns:
            dataset_params["weight"] = features_dataframe.loc[:, "weight"].values
        if target in features_dataframe.columns:
            dataset_params["label"] = features_dataframe.loc[:, target].values
        features_dataframe_casted = xgboost.DMatrix(**dataset_params)
        return features_dataframe_casted
    
    def fit(self, train_features, valid_features, input_features, target, categorical_features, fit_kwargs=dict()):
        train_features_casted = self.cast_dataframe(train_features, input_features, target, categorical_features)
        valid_features_casted = self.cast_dataframe(valid_features, input_features, target, categorical_features) \
                                if valid_features is not None else None
        model_params = dict(self.model_params)
        training_params = {"dtrain":train_features_casted,
                           "num_boost_round":model_params.pop("num_boost_round"),
                           "early_stopping_rounds":model_params.pop("early_stopping_rounds")}
        if valid_features is not None:
            training_params["evals"] = [(valid_features_casted,"eval"),]
        elif "early_stopping_rounds" in training_params:
            del training_params["early_stopping_rounds"]
        training_params["params"] = model_params
        # model training
        self.model = xgboost.train(**training_params, **fit_kwargs)
        self.best_iteration = self.model.best_ntree_limit
        self.input_features = input_features
        self.target = target
        self.categorical_features = categorical_features
    
    def predict(self, predict_features):
        predict_features_casted = self.cast_dataframe(predict_features, self.input_features, self.target, self.categorical_features)
        prediction = self.model.predict(predict_features_casted, ntree_limit=self.best_iteration)
        return prediction

    def save_model(self, fname, **kwargs):
        if self.model is not None:
            self.model.save_model(fname, **kwargs)
    
    def load_model(self, fname, **kwargs):
        self.model = xgboost.Booster()
        self.model.load_model(fname **kwargs)



import abc

class BaseRegressor(abc.ABC):
    @abc.abstractmethod
    def cast_dataframe(self, features_dataframe, input_features, target,  categorical_features):
        """
        Parameters
        ----------
        features_dataframe: pandas.DataFrame
            Dataframe containing all the features.
        input_features: list
            List of column names to use as predictor features.
        target: str
            Name of column to use as target variable.
        categorical_features: list
            List of column names of categorical features.
        Returns
        ----------
        features_dataframe_casted
            Features dataframe casted to the corresponding dataframe format.
        """
        pass
    
    @abc.abstractmethod
    def fit(self, train_features, valid_features, input_features, target, categorical_features):
        """
        Parameters
        ----------
        train_features: pandas.DataFrame
            Dataframe containing the features corresponding to the training data.
        valid_features: pandas.DataFrame
            Dataframe containing the features corresponding to the validation period.
        input_features: list
            List of column names to use as predictor features.
        target: str
            Name of column to use as target variable.
        categorical_features: list
            List of column names of categorical features.
        """
        pass

    @abc.abstractmethod
    def predict(self, predict_features):
        """
        Parameters
        ----------
        predict_features: pandas.DataFrame
            Dataframe containing the features corresponding to the prediction period.
        Returns
        ----------
        predict: numpy.ndarray
            Array with the predicted values.
        """
        pass



import numpy as np
import pandas as pd
import unittest
from parameterized import parameterized_class
from tsforest.forecaster import (CatBoostForecaster,
                                 LightGBMForecaster,
                                 XGBoostForecaster,
                                 H2OGBMForecaster)
from tsforest.utils import make_time_range
from tsforest.trend import compute_trend_models
from tsforest.scaler import compute_scalers

def get_default_model_params(model_class):
    if model_class == CatBoostForecaster:
        return {'iterations':30,'learning_rate':0.3}
    elif model_class == LightGBMForecaster:
        return {'num_iterations':30, 'learning_rate':0.3}
    elif model_class == XGBoostForecaster:
        return {'num_boost_round':30, 'learning_rate':0.3}
    elif model_class == H2OGBMForecaster:
        return {'ntrees':30,'learn_rate':0.3}

TEST_DATA = ['./tests/tests_data/data_single_ts.csv',
             './tests/tests_data/data_many_ts.csv']
TEST_MODELS = [CatBoostForecaster,
               LightGBMForecaster,
               XGBoostForecaster,
               H2OGBMForecaster]

@parameterized_class([
    {"data_path":data_path, "model_class":model_class}
    for model_class in TEST_MODELS
    for data_path in TEST_DATA
])
class TestForecaster(unittest.TestCase):
    def setUp(self):
        data = pd.read_csv(self.data_path, parse_dates=['ds'])
        self.train_data = data.query("ds <= '2019-04-30'")
        self.valid_index = data.query("'2019-04-01' <= ds <= '2019-04-30'").index
        self.eval_data = data.query("'2019-05-01' <= ds <= '2019-05-31'")
        self.predict_data = self.eval_data.drop("y", axis=1)

    def test_it_fit(self):
        model_kwargs = {"model_params":get_default_model_params(self.model_class),
                        "feature_sets":['calendar', 'calendar_cyclical']}
        if "ts_uid" in self.train_data.columns:
            model_kwargs["ts_uid_columns"] = ["ts_uid"]

        fcaster = self.model_class(**model_kwargs)
        fcaster.fit(train_data=self.train_data)
    
    def test_it_fit_with_valid_index(self):
        model_kwargs = {"model_params":get_default_model_params(self.model_class),
                        "feature_sets":['calendar', 'calendar_cyclical']}
        if "ts_uid" in self.train_data.columns:
            model_kwargs["ts_uid_columns"] = ["ts_uid"]

        fcaster = self.model_class(**model_kwargs)
        fcaster.fit(train_data=self.train_data, valid_index=self.valid_index)

    def test_it_fit_with_lag_features(self):
        model_kwargs = {"model_params":get_default_model_params(self.model_class),
                        "feature_sets":['calendar', 'calendar_cyclical', 'lag'],
                        "lags":[1,2,3,4,5,6,7]}
        if "ts_uid" in self.train_data.columns:
            model_kwargs["ts_uid_columns"] = ["ts_uid"]

        fcaster = self.model_class(**model_kwargs)
        fcaster.fit(train_data=self.train_data, valid_index=self.valid_index)

    def test_it_fit_with_rw_features(self):
        model_kwargs = {"model_params":get_default_model_params(self.model_class),
                        "feature_sets":['calendar', 'calendar_cyclical', 'rw'],
                        "window_functions":['mean','median','min','max','sum'],
                        "window_sizes":[7,14,21,28]}
        if "ts_uid" in self.train_data.columns:
            model_kwargs["ts_uid_columns"] = ["ts_uid"]

        fcaster = self.model_class(**model_kwargs)
        fcaster.fit(train_data=self.train_data, valid_index=self.valid_index)

    def test_it_fit_predict(self):
        model_kwargs = {"model_params":get_default_model_params(self.model_class),
                        "feature_sets":['calendar', 'calendar_cyclical']}
        if "ts_uid" in self.train_data.columns:
            model_kwargs["ts_uid_columns"] = ["ts_uid"]

        fcaster = self.model_class(**model_kwargs)
        fcaster.fit(train_data=self.train_data)
        _ = fcaster.predict(self.predict_data)

    def test_it_fit_predict_with_lag_features(self):
        model_kwargs = {"model_params":get_default_model_params(self.model_class),
                        "feature_sets":['calendar', 'calendar_cyclical', 'lag'],
                        "lags":[1,2,3,4,5,6,7]}
        if "ts_uid" in self.train_data.columns:
            model_kwargs["ts_uid_columns"] = ["ts_uid"]

        fcaster = self.model_class(**model_kwargs)
        fcaster.fit(train_data=self.train_data)
        _ = fcaster.predict(self.predict_data)

    def test_it_fit_predict_with_rw_features(self):
        model_kwargs = {"model_params":get_default_model_params(self.model_class),
                        "feature_sets":['calendar', 'calendar_cyclical', 'rw'],
                        "window_functions":['mean','median','min','max','sum'],
                        "window_sizes":[7,14,21,28]}
        if "ts_uid" in self.train_data.columns:
            model_kwargs["ts_uid_columns"] = ["ts_uid"]

        fcaster = self.model_class(**model_kwargs)
        fcaster.fit(train_data=self.train_data)
        _ = fcaster.predict(self.predict_data)
    
    def test_it_fit_predict_with_detrend(self):
        model_kwargs = {"model_params":get_default_model_params(self.model_class),
                        "feature_sets":['calendar', 'calendar_cyclical']}
        if "ts_uid" in self.train_data.columns:
            model_kwargs["ts_uid_columns"] = ["ts_uid"]
            trend_models = compute_trend_models(self.train_data, ts_uid_columns=["ts_uid"])
        else:
            trend_models = compute_trend_models(self.train_data)
        model_kwargs["trend_models"] = trend_models

        fcaster = self.model_class(**model_kwargs)
        fcaster.fit(train_data=self.train_data)
        _ = fcaster.predict(self.predict_data)

    def test_it_fit_predict_with_scaling(self):
        model_kwargs = {"model_params":get_default_model_params(self.model_class),
                        "feature_sets":['calendar', 'calendar_cyclical']}
        if "ts_uid" in self.train_data.columns:
            model_kwargs["ts_uid_columns"] = ["ts_uid"]
            target_scalers = compute_scalers(self.train_data, ts_uid_columns=["ts_uid"])
        else:
            target_scalers = compute_scalers(self.train_data)
        model_kwargs["target_scalers"] = target_scalers
        
        fcaster = self.model_class(**model_kwargs)
        fcaster.fit(train_data=self.train_data)
        _ = fcaster.predict(self.predict_data)

    def test_it_fit_predict_with_detrend_and_scaling(self):
        model_kwargs = {"model_params":get_default_model_params(self.model_class),
                        "feature_sets":['calendar', 'calendar_cyclical']}
        if "ts_uid" in self.train_data.columns:
            model_kwargs["ts_uid_columns"] = ["ts_uid"]
            trend_models = compute_trend_models(self.train_data, ts_uid_columns=["ts_uid"])
            target_scalers = compute_scalers(self.train_data, ts_uid_columns=["ts_uid"])
        else:
            trend_models = compute_trend_models(self.train_data)
            target_scalers = compute_scalers(self.train_data)
        model_kwargs["trend_models"] = trend_models
        model_kwargs["target_scalers"] = target_scalers
        
        fcaster = self.model_class(**model_kwargs)
        fcaster.fit(train_data=self.train_data)
        _ = fcaster.predict(self.predict_data)

    def test_it_fit_evaluate(self):
        model_kwargs = {"model_params":get_default_model_params(self.model_class),
                        "feature_sets":['calendar', 'calendar_cyclical']}
        if "ts_uid" in self.train_data.columns:
            model_kwargs["ts_uid_columns"] = ["ts_uid"]
        
        fcaster = self.model_class(**model_kwargs)
        fcaster.fit(train_data=self.train_data)
        error = fcaster.evaluate(self.eval_data)

        assert type(error)==np.float64, \
            f'fcaster.evaluate returns {error} which is not of type numpy.float64'
    
    def test_it_fit_evaluate_with_bounded_error(self):
        model_kwargs = {"model_params":get_default_model_params(self.model_class),
                        "feature_sets":['calendar', 'calendar_cyclical']}
        if "ts_uid" in self.train_data.columns:
            model_kwargs["ts_uid_columns"] = ["ts_uid"]

        fcaster = self.model_class(**model_kwargs)
        fcaster.fit(train_data=self.train_data)
        error = fcaster.evaluate(self.eval_data)

        assert error <= 3.0, \
            f"fcaster.evaluate returns error={error} which is greater than 3.0"

    def test_it_fit_evaluate_with_lag_features_with_bounded_error(self):
        model_kwargs = {"model_params":get_default_model_params(self.model_class),
                        "feature_sets":['calendar', 'calendar_cyclical', 'lag'],
                        "lags":[1,2,3,4,5,6,7]}
        if "ts_uid" in self.train_data.columns:
            model_kwargs["ts_uid_columns"] = ["ts_uid"]

        fcaster = self.model_class(**model_kwargs)
        fcaster.fit(train_data=self.train_data)
        error = fcaster.evaluate(self.eval_data)

        assert error <= 3.0, \
            f"fcaster.evaluate returns error={error} which is greater than 3.0"

    def test_it_fit_evaluate_with_rw_features_with_bounded_error(self):
        model_kwargs = {"model_params":get_default_model_params(self.model_class),
                        "feature_sets":['calendar', 'calendar_cyclical', 'rw'],
                        "window_functions":['mean','median','min','max','sum'],
                        "window_sizes":[7,14,21,28]}
        if "ts_uid" in self.train_data.columns:
            model_kwargs["ts_uid_columns"] = ["ts_uid"]

        fcaster = self.model_class(**model_kwargs)
        fcaster.fit(train_data=self.train_data)
        error = fcaster.evaluate(self.eval_data)

        assert error <= 3.0, \
            f"fcaster.evaluate returns error={error} which is greater than 3.0"

    def test_it_fit_evaluate_with_detrend_with_bounded_error(self):
        model_kwargs = {"model_params":get_default_model_params(self.model_class),
                        "feature_sets":['calendar', 'calendar_cyclical']}
        if "ts_uid" in self.train_data.columns:
            model_kwargs["ts_uid_columns"] = ["ts_uid"]
            trend_models = compute_trend_models(self.train_data, ts_uid_columns=["ts_uid"])
        else:
            trend_models = compute_trend_models(self.train_data)
        model_kwargs["trend_models"] = trend_models

        fcaster = self.model_class(**model_kwargs)
        fcaster.fit(train_data=self.train_data)
        error = fcaster.evaluate(self.eval_data)

        assert error <= 3.0, \
            f"fcaster.evaluate returns error={error} which is greater than 3.0"

    def test_it_fit_evaluate_with_scaling_with_bounded_error(self):
        model_kwargs = {"model_params":get_default_model_params(self.model_class),
                        "feature_sets":['calendar', 'calendar_cyclical']}
        if "ts_uid" in self.train_data.columns:
            model_kwargs["ts_uid_columns"] = ["ts_uid"]
            target_scalers = compute_scalers(self.train_data, ts_uid_columns=["ts_uid"])
        else:
            target_scalers = compute_scalers(self.train_data)
        model_kwargs["target_scalers"] = target_scalers

        fcaster = self.model_class(**model_kwargs)
        fcaster.fit(train_data=self.train_data)
        error = fcaster.evaluate(self.eval_data)

        assert error <= 3.0, \
            f"fcaster.evaluate returns error={error} which is greater than 3.0"

    def test_it_fit_evaluate_with_detrend_and_scaling_with_bounded_error(self):
        model_kwargs = {"model_params":get_default_model_params(self.model_class),
                        "feature_sets":['calendar', 'calendar_cyclical']}
        if "ts_uid" in self.train_data.columns:
            model_kwargs["ts_uid_columns"] = ["ts_uid"]
            trend_models = compute_trend_models(self.train_data, ts_uid_columns=["ts_uid"])
            target_scalers = compute_scalers(self.train_data, ts_uid_columns=["ts_uid"])
        else:
            trend_models = compute_trend_models(self.train_data)
            target_scalers = compute_scalers(self.train_data)
        model_kwargs["trend_models"] = trend_models
        model_kwargs["target_scalers"] = target_scalers

        fcaster = self.model_class(**model_kwargs)
        fcaster.fit(train_data=self.train_data)
        error = fcaster.evaluate(self.eval_data)

        assert error <= 3.0, \
            f"fcaster.evaluate returns error={error} which is greater than 3.0"



import h2o
from tsforest.forest import H2OGBMRegressor, LightGBMRegressor, CatBoostRegressor, XGBoostRegressor
from tsforest.forecaster_base import ForecasterBase

class H2OGBMForecaster(ForecasterBase):
    def __init__(self, *args, **kwargs):
        super(H2OGBMForecaster, self).__init__(*args, **kwargs)
        # init the cluster if is not already up
        if h2o.cluster() is None: h2o.init(nthreads=-1)
        model_params = kwargs["model_params"] if "model_params" in kwargs else dict()
        self.model = H2OGBMRegressor(model_params)
        for feature,encoding in self.categorical_features.items():
            if encoding == "default":
                self.categorical_features[feature] = "CatBoostEncoder"

class LightGBMForecaster(ForecasterBase):
    def __init__(self, *args, **kwargs):
        super(LightGBMForecaster, self).__init__(*args, **kwargs)
        self.model = LightGBMRegressor(self.model_params)

class CatBoostForecaster(ForecasterBase):
    def __init__(self, *args, **kwargs):
        super(CatBoostForecaster, self).__init__(*args, **kwargs)
        self.model = CatBoostRegressor(self.model_params)

class XGBoostForecaster(ForecasterBase):
    def __init__(self, *args, **kwargs):
        super(XGBoostForecaster, self).__init__(*args, **kwargs)
        self.model = XGBoostRegressor(self.model_params)
        for feature,encoding in self.categorical_features.items():
            if encoding == "default":
                self.categorical_features[feature] = "CatBoostEncoder"



import numpy as np
import pandas as pd
from fbprophet import Prophet
from stldecompose import decompose, forecast
from stldecompose.forecast_funcs import (naive,
                                         drift, 
                                         mean, 
                                         seasonal_naive)
from tsforest.config import prophet_kwargs, stl_kwargs
from joblib import Parallel, delayed
from tqdm import tqdm

import warnings
warnings.filterwarnings("ignore", message="invalid value encountered in subtract")
warnings.simplefilter(action="ignore", category=FutureWarning)
import logging
logging.getLogger("fbprophet").setLevel(logging.WARNING)


def fit_prophet_model(data, kwargs=dict()):
    """
    It receives the dataset and Prophet's model parameters 
    and returns a trained instance of the model
    """ 
    kwargs_extra = dict()
    keys_extra = ["wk_fourier_order", "wk_prior_scale", "mt_fourier_order", 
                  "mt_prior_scale", "yr_fourier_order", "yr_prior_scale"]
    for key in keys_extra:
        if key in kwargs.keys(): kwargs_extra[key] = kwargs.pop(key)

    if len(kwargs) == 0: prophet_model = Prophet()
    else: prophet_model = Prophet(**kwargs) 
        
    if len(kwargs_extra) > 0:
        if "wk_prior_scale" in kwargs_extra:
            prophet_model.add_seasonality(name="weekly", period=7, 
                                          fourier_order=kwargs_extra["wk_fourier_order"],
                                          prior_scale=kwargs_extra["wk_prior_scale"])
        if "mt_prior_scale" in kwargs_extra:
            prophet_model.add_seasonality(name="monthly", period=30.5, 
                                          fourier_order=kwargs_extra["mt_fourier_order"],
                                          prior_scale=kwargs_extra["mt_prior_scale"])
        if "yr_prior_scale" in kwargs_extra:
            prophet_model.add_seasonality(name="yearly", period=365.25, 
                                          fourier_order=kwargs_extra["yr_fourier_order"],
                                          prior_scale=kwargs_extra["yr_prior_scale"])
    # model fitting
    prophet_model.fit(data)    
    return prophet_model

def compute_prophet_trend(prophet_model, predict_dataframe=None, periods=30, include_history=False):
    if predict_dataframe is None:
        predict_dataframe = prophet_model.make_predict_dataframe(periods=periods, 
                                                                 include_history=include_history)
        trend = prophet_model.predict_trend(predict_dataframe)
    else:
        predict_dataframe = prophet_model.setup_dataframe(predict_dataframe.copy())
        trend = prophet_model.predict_trend(predict_dataframe)
    return trend

def fit_stl_model(data, kwargs):
    """
    Parameters
    -----------
    data: pandas.DataFrame 
        trainig data with columns 'ds' (dates) and 'y' (values)
    period: int
        period of the seasonal component to fit
    Returns
    ----------
    pd.DataFrame
        A dataframe with columns 'ds' and 'trend' with the trend
        estimation and projection
    """
    # filling the time gaps
    data = data.copy()
    data.set_index("ds", inplace=True)
    data = (data
            .resample("D")
            .mean()
            .interpolate("linear"))
    # trend estimation with stl
    stl_model = decompose(data, **kwargs)
    return stl_model

def compute_stl_trend(stl_model, predict_dataframe=None):
    """
    Parameters
    -----------
    stl_model: statsmodels.tsa.seasonal.DecomposeResult
        fitted stl model
    predict_dataframe: pandas.DataFrame
        dataframe containg the column 'ds' with times/dates to predict
    Returns
    ----------
    pd.DataFrame
        A dataframe with columns 'ds' and 'trend' with the trend
        estimation and projection
    """
    stl_trend = stl_model.trend.copy()
    stl_trend.rename(columns={"y":"trend"}, inplace=True)
    # calculating the number of predictions to predict
    delta = predict_dataframe.ds.max() - stl_trend.index.max()
    freq = predict_dataframe.ds.diff().min()
    n_periods = int(delta / freq)
    # predicting
    stl_fcst = forecast(stl_model, steps=n_periods, fc_func=drift)
    stl_fcst.rename(columns={"drift":"trend"}, inplace=True)
    stl_trend = stl_trend.append(stl_fcst)
    # keep dates within prediction_dataframe 
    trend = stl_trend.loc[predict_dataframe.ds]
    trend.reset_index(inplace=True)
    trend.rename({"index":"ds"}, axis=1, inplace=True)   
    return trend

class TrendModel():
    """
    Parameters
    ----------
    backend: string
        Name of the backend to be used: 'prophet' or 'stl'.
    """

    def __init__(self, backend="prophet"):
        self.backend = backend

    def fit(self, data, model_kwargs=dict()):
        """
        Parameters
        ----------
        data: pandas.DataFrame
            Dataframe with columns 'ds' and 'y'.
        """
        if self.backend == "prophet":
            trend_model = fit_prophet_model(data, prophet_kwargs)
        elif self.backend == "stl":
            trend_model = fit_stl_model(data, stl_kwargs)
        self.trend_model = trend_model

    def predict(self, predict_period):
        """
        Parameters
        ----------
        predict_period: pandas.DataFrame
            Dataframe with column 'ds' containing the time period to be predicted.
        """
        if self.backend == "prophet":
            trend = compute_prophet_trend(self.trend_model, predict_dataframe=predict_period)
            trend_dataframe = pd.DataFrame({"ds":predict_period.ds.values, "trend":trend.values})
        if self.backend == "stl":
            trend_dataframe = compute_stl_trend(self.trend_model, predict_dataframe=predict_period) 
        return trend_dataframe


def compute_trend_model(train_data_chunk, key):
    if train_data_chunk.shape[0] >= 365:
        trend_model = TrendModel(backend="prophet")
        trend_model.fit(data=train_data_chunk.loc[:, ["ds", "y"]])
    else:
        trend_model = TrendModel(backend="stl")
        trend_model.fit(data=train_data_chunk.loc[:, ["ds", "y"]])
    return {key:trend_model}

def compute_trend_models(data, valid_index=pd.Index([]), ts_uid_columns=None, n_jobs=-1):
    train_data = data.drop(valid_index, axis=0)
    if ts_uid_columns is None:
        train_data["ts_uid"] = 0
        ts_uid_columns = ["ts_uid"]
    ts_uid_values = train_data.loc[:, ts_uid_columns].drop_duplicates()

    train_data_split = list()
    for _,row in ts_uid_values.iterrows():
        key = tuple([item for _,item in row.iteritems()])
        query_string = " & ".join([f"{col_name}=={value}" for col_name,value in row.iteritems()])
        train_data_chunk = train_data.query(query_string)
        train_data_split.append((train_data_chunk, key))
    
    with Parallel(n_jobs=n_jobs) as parallel:
        delayed_func = delayed(compute_trend_model)
        with tqdm(train_data_split) as tqdm_train_data_split:
            trend_models = parallel(delayed_func(*data_chunk)
                                    for data_chunk in tqdm_train_data_split)
        tqdm_train_data_split.close()

    trend_models_out = {}
    for trend_model in trend_models: trend_models_out.update(trend_model)
    return trend_models_out



import numpy as np
import pandas as pd
from tsforest.config import (calendar_features_names,
                             calendar_cyclical_features_names)

def compute_train_features(data, include_features, lags, window_sizes, window_functions, ignore_const_cols=True):
    """
    Parameters
    ----------
    data : pd.DataFrame
        Dataframe with (at least) columns: 'ds' and 'y'.
    include_features: list
        Features to included in computation.
    lags: list
        List of integer lag values.
    window_sizes: list
        List of integer window sizes values.
    window_functions: list
        List of string names of the window functions.
    ignore_const_cols: bool
        Specify whether to ignore constant columns.
    Returns
    ----------
    all_features: pd.Dataframe
        Dataframe containing all the features for the time series.
    """
    # list with all the dataframes of features
    all_features_list = list()

    # generating the calendar features
    if {"calendar","calendar_cyclical"} & set(include_features):
        input_params = {"date_range":pd.DatetimeIndex(data.ds),
                        "ignore_const_cols":ignore_const_cols}
        calendar_features = compute_calendar_features(**input_params)
    if "calendar" not in include_features:
        columns_to_drop = calendar_features_names
        calendar_features.drop(columns=columns_to_drop, inplace=True)
    elif "calendar_cyclical" not in include_features:
        columns_to_drop = calendar_cyclical_features_names
        calendar_features.drop(columns=columns_to_drop, inplace=True)
    all_features_list.append(calendar_features.set_index("ds"))

    # filling time gaps for 'lag' and 'rw' features
    if {"lag", "rw"} & set(include_features):
        filled_data = fill_time_gaps(data)

    if "lag" in include_features:
        lag_features = (compute_lag_features(filled_data, lags=lags)
                        .merge(data.loc[:, ["ds"]], how="inner", on=["ds"]))
        all_features_list.append(lag_features.set_index("ds"))

    if "rw" in include_features:
        rw_features = (compute_rw_features(filled_data, 
                                            window_sizes=window_sizes, 
                                            window_functions=window_functions)
                        .merge(data.loc[:, ["ds"]], how="inner", on=["ds"]))
        all_features_list.append(rw_features.set_index("ds"))

    # merging all features
    all_features_list.append(data.set_index("ds"))
    all_features = (pd.concat(all_features_list, axis=1)
                    .reset_index()
                    .rename({"index":"ds"}, axis=1)
                    .assign(y = lambda x: x["y"].fillna(0.)))
    return all_features

def compute_predict_features(data, include_features, lags, window_sizes, window_functions, ignore_const_cols=True):
    """
    Parameters
    ----------
    data : pd.DataFrame
        Dataframe with (at least) columns: 'ds' and 'y'.
    include_features: list
        Features to included in computation.
    lags: list
        List of integer lag values.
    window_sizes: list
        List of integer window sizes values.
    window_functions: list
        List of string names of the window functions.
    ignore_const_cols: bool
        Specify whether to ignore constant columns.
    Returns
    ----------
    all_features: pd.Dataframe
        Dataframe containing all the features for the time series.
    """
    # list with all the dataframes of features
    all_features_list = list()

    # generating the calendar features
    if np.any(["calendar" in feat for feat in include_features]):
        input_params = {"date_range":pd.DatetimeIndex(data.ds),
                        "ignore_const_cols":ignore_const_cols}
        calendar_features = compute_calendar_features(**input_params)
    if "calendar" not in include_features:
        columns_to_drop = calendar_features_names
        calendar_features.drop(columns=columns_to_drop, inplace=True)
    elif "calendar_cyclical" not in include_features:
        columns_to_drop = calendar_cyclical_features_names
        calendar_features.drop(columns=columns_to_drop, inplace=True)
    all_features_list.append(calendar_features.set_index("ds"))

    if "lag" in include_features:
        column_names = [f"lag_{lag}" for lag in lags]
        lag_features = pd.DataFrame(np.nan, index=data.ds, 
                                    columns=column_names)
        all_features_list.append(lag_features)

    if "rw" in include_features:
        column_names = [f"{window_func}_{window}"
                        for window_func in window_functions
                        for window in window_sizes]
        rw_features = pd.DataFrame(np.nan, index=data.ds,
                                    columns=column_names)
        all_features_list.append(rw_features)
    
    # merging all features
    all_features_list.append(data.set_index("ds"))
    all_features = pd.concat(all_features_list, axis=1)
    all_features.reset_index(inplace=True)
    return all_features

def fill_time_gaps(data, freq="D"):
    """
    Parameters
    ----------
    data: pandas.DataFrame
        Dataframe with columns 'ds' (dtype datetime64) and 'y' 
    """
    assert set(["ds","y"]) <= set(data.columns), "Data must contain the column 'ds'."
    filled_data = (data
                   .resample(freq, on="ds").y.mean()
                   .interpolate("linear")
                   .reset_index())
    filled_data = pd.merge(filled_data, data.drop("y", axis=1), on=["ds"], how="left")
    return filled_data

def compute_calendar_features(date_range, ignore_const_cols=True):
    """
    Parameters
    ----------
    date_range: pandas.DatetimeIndex or pandas.TimedeltaIndex
        Ranges of date times.
    ignore_const_cols: bool
        Specify whether to ignore constant columns.
    """  
    calendar_data = pd.DataFrame()
    calendar_data["ds"] = date_range
    calendar_features = ["year", "quarter", "month", "days_in_month", 
                         "weekofyear", "dayofyear", "day", "dayofweek", 
                         "hour", "minute", "second", "microsecond", 
                         "nanosecond"]
    calendar_features_map = {"weekofyear":"year_week",
                             "dayofyear":"year_day",
                             "day":"month_day",
                             "dayofweek":"week_day"}

    for feature in calendar_features:
        feature_series =  getattr(date_range, feature)
        if feature_series.nunique() == 1 and ignore_const_cols: 
            continue
        feature = calendar_features_map[feature] if feature in calendar_features_map else feature
        calendar_data[feature] = feature_series
    
    # adds missing features
    if (pd.infer_freq(date_range) in ["L", "ms", "U", "us", "N"] 
        and "microsecond" in calendar_data.columns):
        calendar_data['millisecond'] = calendar_data.microsecond//1000
    if {"month_day", "days_in_month"} < set(calendar_data.columns):
        calendar_data["month_progress"] = calendar_data.month_day/calendar_data.days_in_month 
    # adds cyclical encodings
    if "second" in calendar_data.columns:
        calendar_data["second_cos"] = np.cos(calendar_data.second*(2.*np.pi/60))
        calendar_data["second_sin"] = np.sin(calendar_data.second*(2.*np.pi/60))
    if "minute" in calendar_data.columns:
        calendar_data["minute_cos"] = np.cos(calendar_data.minute*(2.*np.pi/60))
        calendar_data["minute_sin"] = np.sin(calendar_data.minute*(2.*np.pi/60))
    if "hour" in calendar_data.columns:
        calendar_data["hour_cos"] = np.cos(calendar_data.hour*(2.*np.pi/24))
        calendar_data["hour_sin"] = np.sin(calendar_data.hour*(2.*np.pi/24))
    if "week_day" in calendar_data.columns:
        calendar_data["week_day_cos"] = np.cos(calendar_data.week_day*(2.*np.pi/7))
        calendar_data["week_day_sin"] = np.sin(calendar_data.week_day*(2.*np.pi/7))
    if "year_day" in calendar_data.columns:
        calendar_data["year_day_cos"] = np.cos((calendar_data.year_day-1)*(2.*np.pi/366))
        calendar_data["year_day_sin"] = np.sin((calendar_data.year_day-1)*(2.*np.pi/366))
    if "year_week" in calendar_data.columns:
        calendar_data["year_week_cos"] = np.cos((calendar_data.year_week-1)*(2.*np.pi/52))
        calendar_data["year_week_sin"] = np.sin((calendar_data.year_week-1)*(2.*np.pi/52))
    if "month" in calendar_data.columns:
        calendar_data["month_cos"] = np.cos((calendar_data.month-1)*(2.*np.pi/12))
        calendar_data["month_sin"] = np.sin((calendar_data.month-1)*(2.*np.pi/12))
    # week_day shifted to 1-7
    if "week_day" in calendar_data.columns:
        calendar_data["week_day"] += 1
    return calendar_data

def compute_lag_features(data, lags):
    """
    data: pandas.Dataframe
        dataframe with column 'y' (response values)
    lags: list
        list of integer lag values
    """
    assert "y" in data.columns, "Missing 'y' column in dataframe."
    features = [data.y.shift(lag) for lag in lags]
    features_names = [f"lag_{lag}" for lag in lags]
    lag_features = pd.concat([data.ds] + features, axis=1)
    lag_features.columns = ["ds"] + features_names
    return lag_features  

def compute_rw_features(data, window_sizes, window_functions):
    """
    data: pandas.Dataframe
        dataframe with column 'y' (response values)
    window_sizes: list
        list of integer window sizes values
    window_functions: list
        list of string names of the window functions
    """
    assert "y" in data.columns, "Missing 'y' column in dataframe"
    # assert window functions in availabe funcs...
    shifted = data.y.shift(1)    
    features = [getattr(shifted.rolling(window), window_func).__call__()
                for window_func in window_functions
                for window in window_sizes]
    features_names = [f"{window_func}_{window}"
                      for window_func in window_functions
                      for window in window_sizes]
    rw_features = pd.concat([data.ds] + features, axis=1)
    rw_features.columns = ["ds"] + features_names
    return rw_features



# h2o gbm default parameters
gbm_parameters = {
    "ntrees":300,
    "max_depth":5,
    "nbins":20,
    "learn_rate":0.25,
    "stopping_rounds":30,
    "stopping_metric":"mse",
    "score_each_iteration":True,
    "categorical_encoding":"enum",
    "sample_rate":1.0,
    "col_sample_rate":1.0,
    "min_rows":5,
    "distribution":"gaussian"
}

# lightgbm default parameters
lgbm_parameters = {
    "boosting_type":"gbrt",
    "objective":"regression",
    "num_iterations":300,
    "num_leaves":31,
    "min_data_in_leaf":5,
    "learning_rate":0.25,
    "feature_fraction":1.0,
    "early_stopping_rounds":30
}

# catboost default parameters
cat_parameters = {
    "iterations":300,
    "learning_rate":0.25,
    "l2_leaf_reg":3.0,
    "depth":6,
    "has_time":True,
    "bootstrap_type":"No",
    "early_stopping_rounds":30
}

# xgboost default parameters
xgb_parameters = {
    "objective":"reg:squarederror",
    "learning_rate":0.25,
    "max_depth":6,
    "lambda":1,
    "num_boost_round":300,
    "early_stopping_rounds":30
}

# fbprophet default parameters 
prophet_kwargs = {
    "n_changepoints":120,
    "changepoint_range":0.9,
    "changepoint_prior_scale":0.05,
    "uncertainty_samples":0,
    "wk_fourier_order" : 3,
    "wk_prior_scale"   : 10.,
    "yr_fourier_order" : 10,
    "yr_prior_scale"   : 10.
} 

# stl default parameters 
stl_kwargs = {
    "period": 7
} 

# names of calendar features
calendar_features_names = [
    "year",
    "quarter",
    "month",
    "days_in_month",
    "year_week",
    "year_day",
    "month_day",
    "week_day",
    "hour",
    "minute",
    "second",
    "millisecond",
    "microsecond",
    "nanosecond",
    "month_progress"
]

# names of cyclical calendar features
calendar_cyclical_features_names = [
    "second_cos",
    "second_sin",
    "minute_cos",
    "minute_sin",
    "hour_cos",
    "hour_sin",
    "week_day_cos",
    "week_day_sin",
    "year_day_cos",
    "year_day_sin",
    "year_week_cos",
    "year_week_sin",
    "month_cos",
    "month_sin"
]



import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from tqdm import tqdm
from sklearn import preprocessing

AVAILABLE_SCALERS = ["MaxAbsScaler", "MinMaxScaler", "Normalizer", 
                     "RobustScaler", "StandardScaler"]

def validate_inputs():
    if target_scaler is not None:
        if not isinstance(target_scaler, str):
            raise TypeError("Parameter 'target_scaler' should be of type 'str'.")
        elif target_scaler not in AVAILABLE_SCALERS:
            raise ValueError(f"Parameter 'target_scaler' should be any of: {AVAILABLE_SCALERS}.")
    
    if not isinstance(target_scaler_kwargs, dict):
        raise TypeError("Parameter 'target_scaler_kwargs' should be of type 'dict'.")


def compute_scaler(train_data_chunk, key, target_scaler, target_scaler_kwargs):
    scaler_class = getattr(preprocessing, target_scaler)
    scaler = scaler_class(**target_scaler_kwargs)
    scaler.fit(train_data_chunk.y.values.reshape(-1,1))
    return {key:scaler}

def compute_scalers(data, valid_index=pd.Index([]), ts_uid_columns=None, 
                    target_scaler="StandardScaler", target_scaler_kwargs=dict(), 
                    n_jobs=-1):
    train_data = data.drop(valid_index, axis=0)
    if ts_uid_columns is None:
        train_data["ts_uid"] = 0
        ts_uid_columns = ["ts_uid"]
    ts_uid_values = train_data.loc[:, ts_uid_columns].drop_duplicates()

    train_data_split = list()
    for _,row in ts_uid_values.iterrows():
        key = tuple([item for _,item in row.iteritems()])
        query_string = " & ".join([f"{col_name}=={value}" for col_name,value in row.iteritems()])
        train_data_chunk = train_data.query(query_string)
        train_data_split.append((train_data_chunk, key))

    with Parallel(n_jobs=n_jobs) as parallel:
        delayed_func = delayed(compute_scaler)
        with tqdm(train_data_split) as tqdm_train_data_split:
            scalers = parallel(delayed_func(*data_chunk, target_scaler=target_scaler, 
                                            target_scaler_kwargs=target_scaler_kwargs)
                               for data_chunk in tqdm_train_data_split)
        tqdm_train_data_split.close()

    scalers_out = {}
    for scaler in scalers: scalers_out.update(scaler)
    return scalers_out



import numpy as np

def compute_mae(yreal, ypred):
    mask = yreal!=0
    yreal = yreal[mask]
    ypred = ypred[mask]
    return np.mean(np.abs(yreal-ypred))

def compute_mse(yreal, ypred):
    mask = yreal!=0
    yreal = yreal[mask]
    ypred = ypred[mask]
    return np.mean((yreal-ypred)**2)

def compute_rmse(yreal, ypred):
    mask = yreal!=0
    yreal = yreal[mask]
    ypred = ypred[mask]
    return np.sqrt(np.mean((yreal-ypred)**2))

def compute_rmsle(yreal, ypred):
    mask = yreal!=0
    yreal = yreal[mask]
    ypred = ypred[mask]
    return np.sqrt(np.mean((np.log(ypred+1) - np.log(yreal+1))**2))
    
def compute_mape(yreal, ypred):
    mask = yreal!=0
    yreal = yreal[mask]
    ypred = ypred[mask]
    return np.mean( np.abs(yreal-ypred)/yreal )

def compute_smape(yreal, ypred):
    mask = yreal!=0
    yreal = yreal[mask]
    ypred = ypred[mask]
    N = yreal.shape[0]
    return np.sum(np.abs(yreal-ypred)/(np.abs(yreal)+np.abs(ypred)))/N



from distutils.core import setup
from setuptools import find_packages

with open('requirements.txt', 'r') as f:
    install_requires = f.read().splitlines()

setup(
    name='tsforest',
    version='0.2.24',
    author='Martín Villanueva',
    author_email='nallivam@gmail.com',
    packages=find_packages(),
    install_requires=install_requires,
    include_package_data=True,
    license='GPLv3',
)



import numpy as np
import pandas as pd
import category_encoders as ce
from inspect import getmembers, isfunction
from tsforest import metrics
from tsforest.trend import TrendModel
from tsforest.features import (compute_train_features, 
                               compute_predict_features)
from tsforest.config import (calendar_features_names,
                             calendar_cyclical_features_names)
from sklearn.base import TransformerMixin

# available methods in pandas.core.window.Rolling as of pandas 0.25.1 
AVAILABLE_RW_FUNCTIONS = ["count", "sum", "mean", "median", "var", "std", "min", 
                          "max", "corr", "cov", "skew", "kurt", "quantile"]

AVAILABLE_METRICS = [member[0].split('_')[1] for member in getmembers(metrics) 
                     if isfunction(member[1])]

AVAILABLE_ENCODERS = [member[1] for member in getmembers(ce)
                      if member[0]=="__all__"][0] + ["default"]

class ForecasterBase(object):
    """
    Parameters
    ----------
    model_params : dict
        Dictionary containing the parameters of the specific boosting model.
    feature_sets: list
        List of feature sets to be included.
    exclude_features: list
        List of features to be excluded from training dataframe.
    categorical_features: dict
        Dict with the name of the categorical feature as keys, and the name
        of the class in 'category_encoders' to be used for encoding as values.
    calendar_anomaly: list
        List of names of calendar features affected by an anomaly.
    ts_uid_columns: list
        List of columns names that are unique identifiers for time series.
    trend_models: dict
        Dictionary with ts_uid as key and TrendModel instances as value.
    target_scalers: dict
        Dictionary with ts_uid as key and TrendModel instances as value.
    lags: list
        List of integer lag values.
    window_sizes: list
        List of integer window sizes values.
    window_functions: list
        List of string names of the window functions.
    """
    def __init__(self, model_params=dict(), feature_sets=["calendar", "calendar_cyclical"], 
                 exclude_features=list(), categorical_features=dict(), calendar_anomaly=list(), 
                 ts_uid_columns=list(), trend_models=dict(), target_scalers=dict(), 
                 lags=None, window_sizes=None, window_functions=None):

        self.model = None
        self.model_params = model_params
        self.feature_sets = feature_sets.copy()
        if lags is not None and "lag" not in feature_sets:
            self.feature_sets.append("lag")
        if (window_sizes is not None and window_functions is not None) and "rw" not in feature_sets:
            self.feature_sets.append("rw")    
        self.exclude_features = ["ds", "y", "y_raw", "weight", "fold_column",
                                 "zero_response", "calendar_anomaly"] + exclude_features
        self.categorical_features = categorical_features.copy()
        self.calendar_anomaly = calendar_anomaly
        self.ts_uid_columns = ts_uid_columns
        for ts_uid_column in ts_uid_columns:  
            if ts_uid_column in categorical_features: continue
            self.categorical_features[ts_uid_column] = "default"
        self.trend_models = trend_models
        self.target_scalers = target_scalers
        self.lags = lags
        self.window_sizes = window_sizes
        self.window_functions = window_functions
        self._features_already_prepared = False
        self._validate_inputs()
    
    def set_params(self, model_params):
        self.model_params = model_params
        self.model.set_params(model_params)
    
    def get_params(self):
        return self.model.get_params()
    
    def _validate_inputs(self):
        """
        Validates the inputs
        """
        if not isinstance(self.model_params, dict):
            raise TypeError("Parameter 'model_params' should be of type 'dict'.")

        if not isinstance(self.feature_sets, list):
            raise TypeError("Parameter 'feature_sets' should be of type 'list'.")
        else:
            if any([x not in ["calendar", "calendar_cyclical", "lag", "rw"] for x in self.feature_sets]):
                raise ValueError("Values in 'feature_sets' should be any of: ['calendar', 'calendar_cyclical', 'lag', 'rw'].")
        
        if not isinstance(self.categorical_features, dict):
            raise TypeError("Parameter 'categorical_features' should be of type 'dict'.")
        else:
            if any([encoding not in AVAILABLE_ENCODERS for encoding in self.categorical_features.values()]):
                raise ValueError(f"Values in 'categorical_features' should be any of: {AVAILABLE_ENCODERS}")

        if not isinstance(self.calendar_anomaly, list):
            raise TypeError("Parameter 'calendar_anomaly' should be of type 'list'.")
        
        if not isinstance(self.ts_uid_columns, list):
            raise TypeError("Parameter 'ts_uid_columns' should be of type 'list'.")

        if not isinstance(self.trend_models, dict):
            raise TypeError("Parameter 'trend_models' should be of type 'dict'.")
        elif not all([isinstance(model, TrendModel) for model in self.trend_models.values()]):
            raise ValueError("Values in 'trend_models' shoud be instances of 'TrendModel'.")
        
        if not isinstance(self.target_scalers, dict):
            raise TypeError("Parameter 'target_scalers' should be of type 'dict'.")
        elif not all([isinstance(scaler, TransformerMixin) for scaler in self.target_scalers.values()]):
            raise ValueError("Values in 'target_scalers' shoud be instances of 'TransformerMixin'.")

        if self.lags is not None:
            if not isinstance(self.lags, list):
                raise TypeError("Parameter 'lags' should be of type 'list'.")
            else:
                if any([type(x)!=int for x in self.lags]):
                    raise ValueError("Values in 'lags' should be integers.")
                elif any([x<1 for x in self.lags]):
                    raise ValueError("Values in 'lags' should be integers greater or equal to 1.")
        
        if self.window_sizes is not None:
            if not isinstance(self.window_sizes, list):
                raise TypeError("Parameter 'window_sizes' should be of type 'list'.")
            else:
                if any([type(x)!=int for x in self.window_sizes]):
                    raise ValueError("Values in 'window_sizes' should be integers.")
                elif any([x<1 for x in self.window_sizes]):
                    raise ValueError("Values in 'window_sizes' should be integers greater or equal to 1.")
        
        if self.window_functions is not None:
            if not isinstance(self.window_functions, list):
                raise TypeError("Parameter 'window_functions' should be of type list.")
            else:
                if any([type(x)!=str for x in self.window_functions]):
                    raise ValueError("Values in 'window_functions' should be string names.")
                elif any([x not in AVAILABLE_RW_FUNCTIONS for x in self.window_functions]):
                    raise ValueError(f"Values in 'window_functions' should be any of: {AVAILABLE_RW_FUNCTIONS}.")
    
    def _validate_input_data(self, train_data, valid_index):
        if not isinstance(train_data, pd.DataFrame):
            raise TypeError("Parameter 'train_data' should be of type pandas.DataFrame.")
        elif not ({"ds", "y"} <= set(train_data.columns.values)):
            raise ValueError("'train_data' should contain columns 'ds' and 'y'.")
        
        if not isinstance(valid_index, pd.Index):
            raise TypeError("Parameter 'valid_index' should be of type 'pandas.Index'.")
        elif not (set(valid_index) <= set(train_data.index)):
            raise ValueError("Parameter 'valid_index' should only contain index values present in 'train_data.index'.") 

        if not set(self.ts_uid_columns) <= set(train_data.columns):
            raise ValueError(f"Parameter 'train_data' has missing ts_uid_columns: {set(self.ts_uid_columns)-set(train_data.columns)}.")
    
    def _validate_predict_data(self, predict_data):
        if not isinstance(predict_data, pd.DataFrame):
            raise TypeError("Parameter 'predict_data' should be of type pandas.DataFrame.")
        elif not (set(self.train_data.columns) - set(predict_data.columns) == {"y", "y_raw"}):
            raise ValueError("'predict_data' shoud have the same columns as 'train_data' except for 'y'.")
    
    def _validate_evaluate_data(self, eval_data, metric):
        if not isinstance(eval_data, pd.DataFrame):
            raise TypeError("'eval_data' should be of type pandas.DataFrame.")
        elif not (set(eval_data.columns) <= set(self.train_data.columns)):
            raise ValueError("'eval_data' should have the same columns as 'train_data'.")

        if not isinstance(metric, str):
            raise TypeError("'metric' should be of type str.")
        elif metric not in AVAILABLE_METRICS:
            raise ValueError(f"'metric' should be any of these: {AVAILABLE_METRICS}")
    
    def _encode_categorical_features(self, train_features, categorical_features, ts_uid_columns):
        categorical_encoders = dict()
        for feature,encoding in categorical_features.items():
            if encoding == "default": 
                if not np.issubdtype(train_features[feature].dtype, np.number):
                    encoding = "OrdinalEncoder"
                else: continue
            encoder_class = getattr(ce, encoding)
            encoder = encoder_class(cols=[feature])
            if feature in ts_uid_columns:
                encoder.fit(train_features.loc[:, [feature]], train_features.loc[:, "y_raw"].values)
            else:
                encoder.fit(train_features.loc[:, [feature]], train_features.loc[:, "y"].values)
            transformed = encoder.transform(train_features.loc[:, [feature]])
            del train_features[feature]
            train_features[transformed.columns] = transformed 
            categorical_encoders[feature] = encoder
        return train_features,categorical_encoders
      
    def _prepare_train_features(self, train_data):
        """
        Parameters
        ----------
        train_data : pandas.DataFrame
            Dataframe with at least columns 'ds' and 'y'.
        """
        train_features = compute_train_features(data=train_data,
                                                include_features=self.feature_sets,
                                                lags=self.lags,
                                                window_sizes=self.window_sizes,
                                                window_functions=self.window_functions,
                                                ignore_const_cols=True)
        if "zero_response" in train_features.columns:
            train_features = train_features.query("zero_response != 1")
        if "calendar_anomaly" in train_features.columns:
            assert len(self.calendar_anomaly) != 0, \
                "'calendar_anomaly' column found, but names of affected features were not provided."
            assert set(self.calendar_anomaly) <= set(train_features.columns), \
                f"Calendar anomaly affected columns: {set(self.calendar_anomaly)-set(train_features.columns)} are not present in 'train_features'."
            idx = train_features.query("calendar_anomaly == 1").index
            train_features.loc[idx, self.calendar_anomaly] = np.nan
        return train_features
    
    def _prepare_predict_features(self, predict_data):
        """
        Parameters
        ----------
        predict_data: pandas.DataFrame
            Dataframe with the same columns as train_data (except for 'y') 
            containing the prediction period.
        Returns
        ----------
        predict_features: pandas.DataFrame
            Dataframe containing all the features for making predictions with 
            the trained model.
        """
        predict_features = compute_predict_features(data=predict_data,
                                                    include_features=self.feature_sets,
                                                    lags=self.lags,
                                                    window_sizes=self.window_sizes,
                                                    window_functions=self.window_functions,
                                                    ignore_const_cols=False)

        if "calendar_anomaly" in predict_features.columns:
            assert len(self.calendar_anomaly) != 0, \
                "'calendar_anomaly' column found, but no names of affected features were provided."
            assert set(self.calendar_anomaly) <= set(self.train_features.columns), \
                f"Calendar anomaly affected columns: {set(self.calendar_anomaly)-set(self.train_features.columns)} are not present in 'train_features'."
            idx = predict_features.query("calendar_anomaly == 1").index
            predict_features.loc[idx, self.calendar_anomaly] = np.nan
        if len(self.categorical_features) > 0:
            for feature,encoder in self.categorical_encoders.items():
                transformed = encoder.transform(predict_features.loc[:, [feature]])
                del predict_features[feature]
                predict_features[transformed.columns] = transformed 

        features_to_keep = [feature for feature in predict_features.columns if feature in self.raw_features]
        return predict_features.loc[:, features_to_keep]
    
    def _prepare_target(self, data, trend_model, target_scaler):
        """
        Prepares the target variable

        Parameters
        ----------
        data: pd.DataFrame
            Dataframe containing the columns 'ds' and 'y'.
        trend_model: TrendModel
            Trained trend estimator model.
        target_scaler: sklearn.base.TransformerMixin
            Trained sklearn.base.TransformerMixin scaler.
        """
        data = data.copy()
        data.y = data.y.astype(float)
        if trend_model is not None:
            trend_dataframe = trend_model.predict(data.loc[:, ["ds"]]) 
            data.loc[:, "y"] -= trend_dataframe.trend.values
        if target_scaler is not None:
            data.loc[:, "y"] = target_scaler.transform(data.y.values.reshape(-1,1)).ravel()
        return data.y.values

    def prepare_train_features(self, train_data, trend_model, target_scaler, sort_by):
        """
        Parameters
        ----------
        train_data : pandas.DataFrame
            Dataframe with at least columns 'ds' and 'y'.
        trend_model: TrendModel
            Trained trend estimator model.
        target_scaler: sklearn.base.TransformerMixin
            Trained sklearn.base.TransformerMixin scaler.
        sort_by: list
            List of column names to sort 'train_data'
        """
        train_data = train_data.sort_values(sort_by, axis=0)
        y_target = self._prepare_target(train_data, trend_model, target_scaler)
        train_data["y_raw"] = train_data.pop("y").values
        train_data["y"] = y_target
        train_features = self._prepare_train_features(train_data)
        train_features.set_index(train_data.index, inplace=True)
        return train_features
    
    def prepare_features(self, train_data, valid_index=pd.Int64Index([])):
        """
        Parameters
        ----------
        train_data : pandas.DataFrame
            Dataframe with at least columns 'ds' and 'y'.
        valid_index: pandas.Index
            Array with indexes from train_data to be used for validation.
        """
        self._validate_input_data(train_data, valid_index)
        # adds a dummy 'ts_uid' column in case of single ts data
        if len(self.ts_uid_columns) == 0: 
            train_data["ts_uid"] = 0
            ts_uid_columns = ["ts_uid"]
        else:
            ts_uid_columns = self.ts_uid_columns

        if (len(self.trend_models) == 0 and
            len(self.target_scalers) == 0 and 
            {"lag", "rw"} & set(self.feature_sets) == set()):
              train_features = self.prepare_train_features(train_data, None, None, sort_by=ts_uid_columns+["ds"])
        else:
            all_train_features = list()
            ts_uid_values = train_data.loc[:, ts_uid_columns].drop_duplicates()
            for _,row in ts_uid_values.iterrows():
                key = tuple([item for _,item in row.iteritems()])
                query_string = " & ".join([f"{col_name}=={value}" for col_name,value in row.iteritems()])
                train_data_chunk = train_data.query(query_string)
                trend_model = self.trend_models[key] if len(self.trend_models) > 0 else None
                target_scaler = self.target_scalers[key] if len(self.target_scalers) > 0 else None
                train_features = self.prepare_train_features(train_data_chunk, trend_model, target_scaler, sort_by=["ds"])
                all_train_features.append(train_features)         
            train_features = pd.concat(all_train_features)

        self.train_data = train_features.loc[:, list(train_data.columns) + ["y_raw"]]
        if len(valid_index) > 0:
            valid_features = train_features.loc[valid_index, :]
            train_features = train_features.drop(valid_index, axis=0)
 
        # performs the encoding of categorical features
        if len(self.categorical_features) > 0:
            train_features,categorical_encoders = self._encode_categorical_features(train_features, self.categorical_features, ts_uid_columns)
            if len(valid_index) > 0:
                for feature,encoder in categorical_encoders.items():
                    transformed = encoder.transform(valid_features.loc[:, [feature]])
                    del valid_features[feature]
                    valid_features[transformed.columns] = transformed 
            self.categorical_encoders = categorical_encoders
        # categorical features to be encoded by the tree/boosting model
        _categorical_features = [feature for feature,encoder in self.categorical_features.items() 
                                 if encoder == "default"]
        self._categorical_features = _categorical_features

        self.raw_features = train_features.columns
        self.input_features = [feature for feature in train_features.columns
                               if feature not in self.exclude_features]
        self.train_features = train_features
        self.valid_features = valid_features if len(valid_index) > 0 else None
        self._features_already_prepared = True
        return self.train_features, self.valid_features

    def fit(self, train_data=None, valid_index=pd.Int64Index([]), fit_kwargs=dict()):
        """
        Parameters
        ----------
        train_data: pandas.DataFrame
            Dataframe with at least columns 'ds' and 'y'.
        valid_index: pandas.Index
            Array with indexes from train_data to be used for validation.
        fit_kwargs: dict
            Extra arguments passed to the fit/train call of the model. 
        """
        if not self._features_already_prepared:
            train_features,valid_features = self.prepare_features(train_data, valid_index)
        else:
            train_features = self.train_features
            valid_features = self.valid_features
        kwargs = {"train_features":train_features, 
                  "valid_features":valid_features,
                  "input_features":self.input_features,
                  "categorical_features":self._categorical_features,
                  "target":"y",
                  "fit_kwargs":fit_kwargs}
        self.model.fit(**kwargs)
        self.best_iteration = self.model.best_iteration

    def predict(self, predict_data):
        """
        Parameters
        ----------
        predict_data: pandas.DataFrame
            Datafame containing the features for the prediction period.
            Contains the same columns as 'train_data' except for 'y'.
        Returns
        ----------
        prediction_dataframe: pandas.DataFrame
            Dataframe containing the dates 'ds' and predictions 'y_pred'.
        """
        # adds a dummy 'ts_uid' column in case of single ts data
        if len(self.ts_uid_columns) == 0: 
            predict_data["ts_uid"] = 0
            ts_uid_columns = ["ts_uid"]
        else:
            ts_uid_columns = self.ts_uid_columns
        self._validate_predict_data(predict_data) 

        if (len(self.trend_models) == 0 and
            len(self.target_scalers) == 0 and
            {"lag", "rw"} & set(self.feature_sets) == set()):
            predict_features = self._prepare_predict_features(predict_data)
            prediction = self.model.predict(predict_features)
            if "zero_response" in predict_features.columns:
                zero_response_mask = predict_features["zero_response"]==1
                prediction[zero_response_mask] = 0
            self.predict_features = predict_features
            prediction_dataframe = (predict_data
                                    .loc[:, ["ds"]+self.ts_uid_columns]
                                    .assign(y_pred = prediction))
        else:
            ts_uid_values = predict_data.loc[:, ts_uid_columns].drop_duplicates()
            all_predict_features = list()
            all_prediction_dataframes = list()
            for _,row in ts_uid_values.iterrows():
                key = tuple([item for _,item in row.iteritems()])
                query_string = " & ".join([f"{col_name}=={value}" for col_name,value in row.iteritems()])
                predict_data_chunk = predict_data.query(query_string)
                predict_features = self._prepare_predict_features(predict_data_chunk)
                if "lag" in self.feature_sets or "rw" in self.feature_sets:
                    y_past = self.train_data.query(query_string).y.values
                    prediction = self._predict(self.model, predict_features, y_past)
                else:
                    prediction = self.model.predict(predict_features)
                
                if len(self.target_scalers) > 0:
                    target_scaler = self.target_scalers[key]
                    prediction = target_scaler.inverse_transform(prediction.reshape(-1,1)).ravel()
                if len(self.trend_models) > 0:
                    trend_model = self.trend_models[key]
                    trend_dataframe = trend_model.predict(predict_data_chunk.loc[:, ["ds"]])
                    prediction += trend_dataframe.trend.values
                if "zero_response" in predict_features.columns:
                    zero_response_mask = predict_features["zero_response"]==1
                    prediction[zero_response_mask] = 0
                
                _prediction_dataframe = (predict_data_chunk
                                         .loc[:, ["ds"]+self.ts_uid_columns]
                                         .assign(y_pred = prediction))
                all_predict_features.append(predict_features)
                all_prediction_dataframes.append(_prediction_dataframe)
            self.predict_features = pd.concat(all_predict_features).reset_index(drop=True)
            prediction_dataframe = pd.concat(all_prediction_dataframes).reset_index(drop=True)
        return prediction_dataframe 

    def _predict(self, model, predict_features, y_past):
        """
        Parameters
        ----------
        model: BaseRegressor
            Instance model of tsforest.forest module.
        predict_features: pandas.DataFrame
            Datafame containing the features for the prediction period.
        y_past: np.ndarray
            Array with the values of 'y' previous to the prediction period.
        """
        y = y_past.tolist()
        for idx in predict_features.index:
            if "lag" in self.feature_sets:
                for lag in self.lags:
                    predict_features.loc[idx, f"lag_{lag}"] = y[-lag]
            if "rw" in self.feature_sets:
                for window_func in self.window_functions:
                    for window in self.window_sizes:
                        predict_features.loc[idx, f"{window_func}_{window}"] = getattr(np, window_func)(y[-window:])
            y_pred = model.predict(predict_features.loc[[idx], self.input_features])
            y.append(y_pred)
        n_predictions = predict_features.shape[0]
        return np.asarray(y[-n_predictions:])

    def evaluate(self, eval_data, metric="rmse"):
        """
        Parameters
        ----------
        eval_data: pandas.DataFrame
            Dataframe with the same columns as 'train_data'.
        metric: string
            Possible values: 'mae', 'mape', 'mse', 'rmse', 'smape'.
        Returns
        ----------
        error: float
            error of predictions according to the error measure
        """
        self._validate_evaluate_data(eval_data, metric)
        eval_data = eval_data.copy()
        y_real = eval_data.pop("y")
        y_pred = self.predict(eval_data)["y_pred"].values
        error_func = getattr(metrics, f"compute_{metric}")
        error = error_func(y_real, y_pred)
        return error

    def save_model(self, fname, **kwargs):
        if self.model is not None:
            self.model.save_model(fname, **kwargs)

    def load_model(self, fname, **kwargs):
        self.model.load_model(fname, **kwargs)
